@@ -20,6 +20,7 @@
  */
 
 #include "gtksourcestyleschememanager.h"
+#include "gtksourcestylescheme.h"
 #include "gtksourceview-marshal.h"
 #include "gtksourceview-i18n.h"
 #include "gtksourceview-utils.h"
@@ -27,9 +28,11 @@
 
 /**
  * SECTION:styleschememanager
- * @Short_description: Object which provides access to #GtkSourceStyleScheme<!-- -->s
+ * @Short_description: Object which provides access to GtkSourceStyleSchemes
  * @Title: GtkSourceStyleSchemeManager
  * @See_also: #GtkSourceStyleScheme
+ *
+ * Object which provides access to #GtkSourceStyleScheme<!-- -->s.
  */
 
 #define SCHEME_FILE_SUFFIX	".xml"
